@@ -1,10 +1,13 @@
 from graphistry.Plottable import Plottable
 from typing import Any, Callable, List, Optional, Union, TYPE_CHECKING
-import copy, hashlib, logging, numpy as np, pandas as pd, pyarrow as pa, sys, uuid
+import copy, hashlib, numpy as np, pandas as pd, pyarrow as pa, sys, uuid
 from functools import lru_cache
 from weakref import WeakValueDictionary
 
-from .util import (error, in_ipython, make_iframe, random_string, warn)
+from .util import (
+    error, hash_pdf, in_ipython, in_databricks, make_iframe, random_string, warn,
+    cache_coercion, cache_coercion_helper, WeakValueWrapper
+)
 
 from .bolt_util import (
     bolt_graph_to_edges_dataframe,
@@ -17,6 +20,8 @@ from .bolt_util import (
 from .arrow_uploader import ArrowUploader
 from .nodexlistry import NodeXLGraphistry
 from .tigeristry import Tigeristry
+from .util import setup_logger
+logger = setup_logger(__name__)
 
 maybe_cudf = None
 try:
@@ -24,6 +29,8 @@ try:
     maybe_cudf = cudf
 except ImportError:
     1
+except RuntimeError:
+    logger.warning('Runtime error import cudf: Available but failed to initialize', exc_info=True)
 
 maybe_dask_dataframe = None
 try:
@@ -38,30 +45,16 @@ try:
     maybe_dask_cudf = dask_cudf
 except ImportError:
     1
+except RuntimeError:
+    logger.warning('Runtime error import dask_cudf: Available but failed to initialize', exc_info=True)
 
-logger = logging.getLogger('Plotter')
+maybe_spark = None
+try:
+    import pyspark
+    maybe_spark = pyspark
+except ImportError:
+    1
 
-CACHE_COERCION_SIZE = 100
-
-
-_cache_coercion_val = None
-@lru_cache(maxsize=CACHE_COERCION_SIZE)
-def cache_coercion_helper(k):
-    return _cache_coercion_val
-
-def cache_coercion(k, v):
-    """
-        Holds references to last 100 used coercions
-        Use with weak key/value dictionaries for actual lookups
-    """
-    global _cache_coercion_val
-    _cache_coercion_val = v
-
-    return cache_coercion_helper(k)
-
-class WeakValueWrapper:
-    def __init__(self, v):
-        self.v = v
 
 class PlotterBase(Plottable):
     """Graph plotting class.
@@ -80,6 +73,17 @@ class PlotterBase(Plottable):
     _defaultNodeId = '__nodeid__'
     _pd_hash_to_arrow : WeakValueDictionary = WeakValueDictionary()
     _cudf_hash_to_arrow : WeakValueDictionary = WeakValueDictionary()
+    _umap_param_to_g : WeakValueDictionary = WeakValueDictionary()
+    _feat_param_to_g : WeakValueDictionary = WeakValueDictionary()
+
+    def reset_caches(self): 
+        """Reset memoization caches"""
+        self._pd_hash_to_arrow.clear()
+        self._cudf_hash_to_arrow.clear()
+        self._umap_param_to_g.clear()
+        self._feat_param_to_g.clear()
+        cache_coercion_helper.cache_clear()
+
 
     def __init__(self, *args, **kwargs):
         super(PlotterBase, self).__init__()
@@ -90,6 +94,7 @@ class PlotterBase(Plottable):
         self._source : Optional[str] = None
         self._destination : Optional[str] = None
         self._node : Optional[str] = None
+        self._edge : Optional[str] = None
         self._edge_title : Optional[str] = None
         self._edge_label : Optional[str] = None
         self._edge_color : Optional[str] = None
@@ -125,6 +130,34 @@ class PlotterBase(Plottable):
         self._bolt_driver : any = None
         self._tigergraph : any = None
 
+        self._node_embedding = None
+        self._node_encoder = None
+        self._node_features = None
+        self._node_imputer = None
+        self._node_scaler = None
+        self._node_target = None
+        self._node_target_encoder = None
+
+        self._edge_embedding = None
+        self._edge_encoders = None
+        self._edge_features = None
+        self._edge_imputer = None
+        self._edge_scaler = None
+        self._edge_target = None
+        self._edge_target_encoder = None
+
+        self._weighted_adjacency_nodes = None
+        self._weighted_adjacency_edges = None
+        self._weighted_edges_df = None
+        self._weighted_edges_df_from_nodes = None
+        self._weighted_edges_df_from_edges = None
+
+        self._umap = None
+
+        self._adjacency = None
+        self._entity_to_index = None
+        self._index_to_entity = None
+
 
     def __repr__(self):
         bindings = ['edges', 'nodes', 'source', 'destination', 'node', 
@@ -254,6 +287,63 @@ class PlotterBase(Plottable):
         return res
 
 
+    def encode_axis(self, rows=[]):
+        """Render radial and linear axes with optional labels
+
+        :param rows: List of rows - {
+            label: Optional[str],
+            ?r: float,
+            ?x: float,
+            ?y: float,
+            ?internal: true,
+            ?external: true,
+            ?space: true
+        }
+
+        :returns: Plotter
+        :rtype: Plotter
+
+        **Example: Several radial axes**
+            ::
+
+                g.encode_axis([
+                  {'r': 14, 'external': True, 'label': 'outermost'},
+                  {'r': 12, 'external': True},
+                  {'r': 10, 'space': True},
+                  {'r': 8, 'space': True},
+                  {'r': 6, 'internal': True},
+                  {'r': 4, 'space': True},
+                  {'r': 2, 'space': True, 'label': 'innermost'}
+                ])
+
+        **Example: Several horizontal axes**
+            ::
+
+                g.encode_axis([
+                  {"label": "a",  "y": 2, "internal": True },
+                  {"label": "b",  "y": 40, "external": True, "width": 20, "bounds": {"min": 40, "max": 400}},
+                ])
+
+        """
+
+        complex_encodings = self._complex_encodings or {}
+        if 'current' not in complex_encodings:
+            complex_encodings['current'] = {}
+        if 'default' not in complex_encodings:
+            complex_encodings['default'] = {}
+        complex_encodings['default']["pointAxisEncoding"] = {
+            "graphType": "point",
+            "encodingType": "axis",
+            "variation": "categorical",
+            "attribute": "degree",
+            "rows": rows
+        }
+
+        out = self.bind()
+        out._complex_encodings = complex_encodings
+        return out
+
+
     def encode_point_color(self, column,
             palette=None, as_categorical=None, as_continuous=None, categorical_mapping=None, default_mapping=None,
             for_default=True, for_current=False):
@@ -686,7 +776,7 @@ class PlotterBase(Plottable):
         return res
 
 
-    def bind(self, source=None, destination=None, node=None,
+    def bind(self, source=None, destination=None, node=None, edge=None,
              edge_title=None, edge_label=None, edge_color=None, edge_weight=None, edge_size=None, edge_opacity=None, edge_icon=None,
              edge_source_color=None, edge_destination_color=None,
              point_title=None, point_label=None, point_color=None, point_weight=None, point_size=None, point_opacity=None, point_icon=None,
@@ -705,6 +795,9 @@ class PlotterBase(Plottable):
         :param node: Attribute containing a node's ID
         :type node: str
 
+        :param edge: Attribute containing an edge's ID
+        :type edge: str
+
         :param edge_title: Attribute overriding edge's minimized label text. By default, the edge source and destination is used.
         :type edge_title: str
 
@@ -791,6 +884,7 @@ class PlotterBase(Plottable):
         res._source = source or self._source
         res._destination = destination or self._destination
         res._node = node or self._node
+        res._edge = edge or self._edge
 
         res._edge_title = edge_title or self._edge_title
         res._edge_label = edge_label or self._edge_label
@@ -814,6 +908,9 @@ class PlotterBase(Plottable):
         
         return res
 
+    def copy(self) -> Plottable:
+        return copy.copy(self)
+
 
     def nodes(self, nodes: Union[Callable, Any], node=None, *args, **kwargs) -> Plottable:
         """Specify the set of nodes and associated data.
@@ -903,7 +1000,7 @@ class PlotterBase(Plottable):
         return res
 
 
-    def edges(self, edges: Union[Callable, Any], source=None, destination=None, *args, **kwargs) -> Plottable:
+    def edges(self, edges: Union[Callable, Any], source=None, destination=None, edge=None, *args, **kwargs) -> Plottable:
         """Specify edge list data and associated edge attribute values.
         If a callable, will be called with current Plotter and whatever positional+named arguments
 
@@ -923,6 +1020,16 @@ class PlotterBase(Plottable):
                     .edges(df)
                     .plot()
 
+        **Example**
+            ::
+
+                import graphistry
+                df = pandas.DataFrame({'src': [0,1,2], 'dst': [1,2,0], 'id': [0, 1, 2]})
+                graphistry
+                    .bind(source='src', destination='dst', edge='id')
+                    .edges(df)
+                    .plot()
+
         **Example**
             ::
 
@@ -932,6 +1039,15 @@ class PlotterBase(Plottable):
                     .edges(df, 'src', 'dst')
                     .plot()
 
+        **Example**
+            ::
+
+                import graphistry
+                df = pandas.DataFrame({'src': [0,1,2], 'dst': [1,2,0], 'id': [0, 1, 2]})
+                graphistry
+                    .edges(df, 'src', 'dst', 'id')
+                    .plot()
+
         **Example**
             ::
                 import graphistry
@@ -944,7 +1060,7 @@ class PlotterBase(Plottable):
                 graphistry
                     .edges(df, 'src', 'dst')
                     .edges(sample_edges, n=2)
-                    .edges(sample_edges, None, None, 2)  # equivalent
+                    .edges(sample_edges, None, None, None, 2)  # equivalent
                     .plot()
 
         """
@@ -955,6 +1071,8 @@ class PlotterBase(Plottable):
             base = base.bind(source=source)
         if not (destination is None):
             base = base.bind(destination=destination)
+        if edge is not None:
+            base = base.bind(edge=edge)
 
         if callable(edges):
             edges2 = edges(base, *args, **kwargs)
@@ -1135,7 +1253,10 @@ class PlotterBase(Plottable):
         return res
 
 
-    def plot(self, graph=None, nodes=None, name=None, description=None, render=None, skip_upload=False, as_files=False, memoize=True):  # noqa: C901
+    def plot(
+        self, graph=None, nodes=None, name=None, description=None, render=None, skip_upload=False, as_files=False, memoize=True,
+        extra_html="", override_html_style=None
+    ):  # noqa: C901
         """Upload data to the Graphistry server and show as an iframe of it.
 
         Uses the currently bound schema structure and visual encodings.
@@ -1167,6 +1288,12 @@ class PlotterBase(Plottable):
         :param memoize: Tries to memoize pandas/cudf->arrow conversion, including skipping upload. Default on.
         :type memoize: bool
 
+        :param extra_html: Allow injecting arbitrary HTML into the visualization iframe.
+        :type extra_html: Optional[str]
+
+        :param override_html_style: Set fully custom style tag.
+        :type override_html_style: Optional[str]
+
         **Example: Simple**
             ::
 
@@ -1234,7 +1361,9 @@ class PlotterBase(Plottable):
             return full_url
         elif (render is True) or in_ipython():
             from IPython.core.display import HTML
-            return HTML(make_iframe(full_url, self._height))
+            return HTML(make_iframe(full_url, self._height, extra_html=extra_html, override_html_style=override_html_style))
+        elif in_databricks():
+            return make_iframe(full_url, self._height, extra_html=extra_html, override_html_style=override_html_style)
         else:
             import webbrowser
             webbrowser.open(full_url)
@@ -1377,7 +1506,8 @@ class PlotterBase(Plottable):
                 or isinstance(graph, pa.Table) \
                 or ( not (maybe_cudf is None) and isinstance(graph, maybe_cudf.DataFrame) ) \
                 or ( not (maybe_dask_cudf is None) and isinstance(graph, maybe_dask_cudf.DataFrame) ) \
-                or ( not (maybe_dask_dataframe is None) and isinstance(graph, maybe_dask_dataframe.DataFrame) ):
+                or ( not (maybe_dask_dataframe is None) and isinstance(graph, maybe_dask_dataframe.DataFrame) ) \
+                or ( not (maybe_spark is None) and isinstance(graph, pyspark.sql.dataframe.DataFrame) ):
             return g._make_dataset(graph, nodes, name, description, mode, metadata, memoize)
 
         try:
@@ -1399,7 +1529,7 @@ class PlotterBase(Plottable):
         except ImportError:
             pass
 
-        error('Expected Pandas/Arrow/cuDF dataframe(s) or Igraph/NetworkX graph.')
+        error('Expected Pandas/Arrow/cuDF/Spark dataframe(s) or igraph/NetworkX graph.')
 
 
     # Sanitize node/edge dataframe by
@@ -1578,10 +1708,9 @@ class PlotterBase(Plottable):
             if memoize:
                 
                 try:
-                    #https://stackoverflow.com/questions/31567401/get-the-same-hash-value-for-a-pandas-dataframe-each-time
-                    hashed = hashlib.sha256(pd.util.hash_pandas_object(table, index=True).values).hexdigest()
+                    hashed = hash_pdf(table)
                 except TypeError:
-                    logger.warn('Failed memoization speedup attempt due to Pandas internal hash function failing. Continuing without memoization speedups.'
+                    logger.warning('Failed memoization speedup attempt due to Pandas internal hash function failing. Continuing without memoization speedups.'
                                 'This is fine, but for speedups around skipping re-uploads of previously seen tables, '
                                 'try identifying which columns have types that Pandas cannot hash, and convert them '
                                 'to hashable types like strings.')
@@ -1610,7 +1739,10 @@ class PlotterBase(Plottable):
             hashed = None
             if memoize:
                 #https://stackoverflow.com/questions/31567401/get-the-same-hash-value-for-a-pandas-dataframe-each-time
-                hashed = hashlib.sha256(table.hash_columns().tobytes()).hexdigest()
+                hashed = (
+                    hashlib.sha256(table.hash_columns().tobytes()).hexdigest()
+                    + hashlib.sha256(str(table.columns).encode('utf-8')).hexdigest()  # noqa: W503
+                )
                 try:
                     if hashed in PlotterBase._cudf_hash_to_arrow:
                         logger.debug('cudf->arrow memoization hit: %s', hashed)
@@ -1643,6 +1775,12 @@ class PlotterBase(Plottable):
             df = ddf.compute()
             return self._table_to_arrow(df, memoize)
 
+        if not (maybe_spark is None) and isinstance(table, maybe_spark.sql.dataframe.DataFrame):
+            logger.debug('spark->arrow via df')
+            df = table.toPandas()
+            #TODO push the hash check to Spark
+            return self._table_to_arrow(df, memoize)
+
         raise Exception('Unknown type %s: Could not convert data to Arrow' % str(type(table)))
 
 
@@ -1690,8 +1828,16 @@ class PlotterBase(Plottable):
 
         from .pygraphistry import PyGraphistry
 
+        def flatten_categorical(df):
+            # Avoid cat_col.where(...)-related exceptions
+            df2 = df.copy()
+            for c in df:
+                if (df[c].dtype.name == 'category'):
+                    df2[c] = df[c].astype(df[c].cat.categories.dtype)
+            return df2
+
         (elist, nlist) = self._bind_attributes_v1(edges, nodes)
-        edict = elist.where((pd.notnull(elist)), None).to_dict(orient='records')
+        edict = flatten_categorical(elist).where(pd.notnull(elist), None).to_dict(orient='records')
 
         bindings = {'idField': self._node or PlotterBase._defaultNodeId,
                     'destinationField': self._destination, 'sourceField': self._source}
@@ -1699,7 +1845,7 @@ class PlotterBase(Plottable):
                    'bindings': bindings, 'type': 'edgelist', 'graph': edict}
 
         if nlist is not None:
-            ndict = nlist.where((pd.notnull(nlist)), None).to_dict(orient='records')
+            ndict = flatten_categorical(nlist).where(pd.notnull(nlist), None).to_dict(orient='records')
             dataset['labels'] = ndict
         return dataset
 
